import warnings

from sqlalchemy import create_engine, text


class Singleton(type):
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class Connection(metaclass=Singleton):
    """Wraps a connection to a MySQL instance."""
    def __init__(self, db_engine_url):
        """Create a connection engine with default connection pooling.

        As a Singleton, the connection engine will only be created once for the entire process.
        SqlAlchemy will then chekout a new connection from the pool as needed.
        """
        self.engine = create_engine(
            db_engine_url,
            max_overflow=2,
            pool_pre_ping=True,
            pool_recycle=1800,  # 30 minutes
            pool_size=5,
            pool_timeout=30
        )

    def execute(self, *args, **kwargs):
        connection = self.engine.connect()
        connection.execute('SET NAMES utf8mb4 COLLATE utf8mb4_unicode_ci')
        connection.execute(*args, **kwargs)
        connection.close()

    def has_table(self, table_name):
        return self.engine.dialect.has_table(self.engine, table_name)

    def query(self, statement, scalar=False, total=None):
        """Run a SQL statement and return the results.

        When a 'total' SQL statement is provided, the total scalar is also returned, and the
        return object is then a tuple.

        When 'scalar' is True, a scalar type is returned.
        """
        connection = self.engine.connect()
        connection.execute('SET NAMES utf8mb4 COLLATE utf8mb4_unicode_ci')

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', '.*Division by 0.*', category=Warning)
            result = connection.execute(text(statement))

        if total:
            totals = connection.execute(total).scalar()

        if scalar:
            data = result.scalar()
        else:
            data = result.fetchall()

        result.close()
        connection.close()

        if total:
            return data, totals

        return data



"""Common RBX exceptions."""


class FatalException(Exception):
    """Raise this exception for non-transient errors.

    These exceptions can be given an extra details object, which may be used by the caller to log
    this as extra information.
    """
    message = 'Something went wrong!'

    def __init__(self, message=None, details=None):
        super().__init__(message or self.message)
        self.details = details

    def to_dict(self):
        rv = {
            'message': str(self)
        }

        if self.details:
            rv['details'] = self.details

        return rv


class HTTPException(FatalException):
    """A FatalException with an HTTP status code and a URL."""
    def __init__(self, message, status_code=400, url=None, **kwargs):
        super().__init__(message, **kwargs)
        self.status_code = status_code
        self.url = url

    def to_dict(self):
        rv = super().to_dict()
        if self.url:
            rv['url'] = self.url

        return rv


class BadRequest(HTTPException):
    """Convenience Exception to use from within a Flask service/blueprint.

    Raising this exception will be caught by the error handler and formatted as an HTTP response,
    using the provided status_code.
    """


class ClientException(HTTPException):
    """Raised from within the rbx.clients package, caused by third-party APIs."""


class TransientException(Exception):
    """Raise this exception when we know the cause is transient (e.g.: connection errors,
    consistency error, ...).
    """
